#!/usr/bin/python
#
# Tools for generating reports

import urllib, re, sys, optparse, os.path, datetime
import sqlite3 as dbapi2
from cgi import escape

valid_states = ['todo', 'done', 'postponed']
blueprints_base_url = 'https://blueprints.launchpad.net'

def escape_sql(text):
    return text.replace("'", "''")

def fix_stdouterr():
    '''Set stdout/stderr encoding.

    Avoid having to do .encode('UTF-8') everywhere. This is a pain; I wish
    Python supported something like "sys.stdout.encoding = 'UTF-8'".
    '''
    import codecs
    def null_decode(input, errors='strict'):
        return input, len(input)
    sys.stdout = codecs.EncodedFile(sys.stdout, 'UTF-8')
    sys.stdout.decode = null_decode
    sys.stderr = codecs.EncodedFile(sys.stderr, 'UTF-8')
    sys.stderr.decode = null_decode

def get_db(dbpath):
    '''Open database.'''

    assert os.path.exists(dbpath)
    return dbapi2.connect(dbpath)

def load_config(config_path):
    '''Load a configuration file and return it as a dictionary.'''

    cfg = {}
    execfile(config_path, cfg)
    del cfg['__builtins__']
    return cfg

def escape_url(url):
    '''Correctly escape data intended to be included in a URL'''

    return urllib.quote(url)

def escape_html(html):
    '''Correctly escape data intended to be included as raw HTML'''

    return escape(html, True)

def workitems_over_time(db, team=None, milestone=None):
    '''Calculate work item development over time.

    If team is given, this filters out blueprints for that particular team. If
    team_assignees_only is True, only work items for that team's members are
    counted, otherwise this counts all work items for specs belonging to the
    given team.

    Return date -> state -> count mapping. states are
    {todo,done,postponed}{,_teamonly}.
    '''
    data = {}
    if milestone:
        ms_sql = "AND w.milestone='%s'" % escape_sql(milestone)
    else:
        ms_sql = ''

    # include both direct team assignments, as well as assignmets to
    # members of that team
    for s in valid_states:
        cur = db.cursor()
        if team:
            # WIs which are assigned to team members
            cur.execute('SELECT date, COUNT(DISTINCT w.description) '
                    'FROM work_items w LEFT JOIN teams '
                    'ON w.assignee = teams.name '
                    'WHERE status = ? AND (team = ? OR assignee = ?) %s'
                    'GROUP BY date' % ms_sql, (s, team, team))

            for (date, num) in cur:
                data.setdefault(date, {})[s + '_teamonly'] = num

            # all WIs which belong to that team's specs
            cur.execute('SELECT date, COUNT(DISTINCT w.description) '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                    'WHERE w.status = ? AND (teams.team = ? OR w.assignee = ? OR s.assignee = ?) %s'
                    'GROUP BY date' % ms_sql, (s, team, team, team))

            for (date, num) in cur:
                data.setdefault(date, {})[s] = num
        else:
            # all WIs for this milestone
            cur.execute('SELECT date, count(*) FROM work_items w '
                    'WHERE status=?  %s GROUP BY date' % ms_sql, (s,))

            for (date, num) in cur:
                data.setdefault(date, {})[s] = num
    return data

def blueprint_completion(db, team=None, milestone=None):
    '''Determine current blueprint completion.

    Return blueprint -> info mapping, with info being a map with these
    keys: todo, done, postponed, status, priority, implementation, url.
    '''
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % escape_sql(milestone)
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    for s in valid_states:
        cur = db.cursor()
        if team:
            # include both direct team assignments, as well as assignmnets to
            # members of that team
            cur.execute('SELECT s.name, COUNT(DISTINCT w.description), s.status, s.priority, s.implementation, s.url '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name)'
                    'WHERE w.status = ? AND w.date = ? AND'
                    '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s '
                    'GROUP BY w.spec' % ms_sql,
                    (s, last_date, team, team, team))
        else:
            cur.execute('SELECT s.name, COUNT(DISTINCT w.description), s.status, s.priority, s.implementation, s.url '
                    'FROM work_items w, specs s '
                    'ON w.spec = s.name '
                    'WHERE w.status = ? AND w.date = ? %s '
                    'GROUP BY w.spec' % ms_sql,
                    (s, last_date))

        for (bp, num, status, priority, impl, url) in cur:
            info = data.setdefault(bp, {'todo': 0, 'done': 0, 'postponed': 0})
            info[s] = num
            info['status'] = status or ''
            info['priority'] = priority
            info['implementation'] = impl
            info['url'] = url

    return data


def blueprint_tasks_completion(db, team=None, milestone=None):
    '''Determine current blueprint tasks completion.

    Return blueprint -> info mapping, with info being a map with these
    keys: todo, done, postponed, status, priority, implementation, url, tasks.
    '''
    default = blueprint_completion(db, team=team)
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % escape_sql(milestone)
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    for s in valid_states:
        cur = db.cursor()
        if team:
            # include both direct team assignments, as well as assignmnets to
            # members of that team
            cur.execute('SELECT DISTINCT w.spec, w.description, w.assignee, w.milestone, w.status '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                    'WHERE w.status = ? AND w.date = ? AND'
                    '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s ' % ms_sql,
                    (s, last_date, team, team, team))
        else:
            cur.execute('SELECT DISTINCT w.spec, w.description, w.assignee, w.milestone, w.status '
                    'FROM work_items w, specs s '
                    'ON w.spec = s.name '
                    'WHERE w.status = ? AND w.date = ? %s '
                    'GROUP BY w.spec' % ms_sql,
                    (s, last_date))

        for (bp, desc, assignee, milestone, status) in cur:
            data.setdefault(bp, default[bp])
            info = data[bp].setdefault('tasks', [])
            info.append({
                'description': desc,
                'assignee': assignee,
                'milestone': milestone,
                'state': status
            })

    return data


def assignee_completion(db, team=None, milestone=None):
    '''Determine current by-assignee completion.

    Return assignee -> info mapping with info being a map with these
    keys: todo, done, postponed. Each of those values is a list of [blueprint,
    workitem, priority, spec_url].
    '''
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % milestone
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    for s in valid_states:
        cur = db.cursor()
        if team:
            # include both direct team assignments, as well as assignmnets to
            # members of that team
            cur.execute('SELECT DISTINCT w.assignee, w.spec, w.description, s.priority, s.url '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                    'WHERE w.status = ? AND w.date = ? AND'
                    '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s ' % ms_sql,
                    (s, last_date, team, team, team))
        else:
            cur.execute('SELECT DISTINCT w.assignee, w.spec, w.description, s.priority, s.url '
                    'FROM work_items w, specs s '
                    'ON w.spec = s.name '
                    'WHERE w.status = ? AND w.date = ? %s ' % ms_sql,
                    (s, last_date))
        for (a, bp, description, priority, url) in cur:
            info = data.setdefault(a, {'todo': [], 'done': [], 'postponed': []})
            info[s].append([bp, description, priority, url])

    return data


def milestone_list(db):
        milestones = []

        cur = db.cursor()
        cur.execute('SELECT m.name FROM milestones m ORDER BY m.due_date')

        for (milestone,) in cur:
            milestones.append(milestone)

        return milestones


def priority_value(name):
    '''Translate a priority string into a sortable numeric value.'''

    trans = {'Essential': 99, 'High': 75, 'Medium': 50, 'Low': 25}

    return trans.get(name, 0)

if __name__ == '__main__':
    # some test stuff
    db = get_db(sys.argv[1])

    print '==== Assignee completion ==='
    for date, data in assignee_completion(db).iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, milestone='lucid-alpha-2').iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, 'canonical-desktop-team').iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, 'canonical-desktop-team', milestone='lucid-alpha-2').iteritems():
        print date, [len(data[x]) for x in data]
